/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Duesterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2002-2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Karsten Duesterloh <mnenhy@tprac.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// +-
// |  the contents of this file will be loaded into the Mnenhy.uninstall.* context!
// |  ie. foo() is available as goMnenhy.uninstall.foo()
// +-

// |
// |  constants
// |

// debug
const gksUninstallDebugPref      = "extensions.mnenhy.debug.uninstall";
const gksUninstallNoBackupPref   = "extensions.mnenhy.uninstall.nobackup";
const gksUninstallNoShutdownPref = "extensions.mnenhy.noshutdown";


// |  CIDs
const gksCID_Fastload            = "@mozilla.org/fast-load-service;1";


// |  installed-chrome.txt entries
// package types
const gksTypeContent             = "content";
const gksTypeSkin                = "skin";
const gksTypeLocale              = "locale";
// package location             
const gksLocInstall              = "install";
const gksLocProfile              = "profile";
const gksLocSelect               = "select";   // skin and locale only; usually once per i-c.txt
// package path uri type        
const gksURIURL                  = "url";      // [jar:]resource: or [jar:]file:
const gksURIPath                 = "path";     // local path, eg. F:\\foo\\bar
// line ends                    
const gksCR                      = "\r";
const gksLF                      = "\n";
const gksCRLF                    = gksCR + gksLF;
                                
                                
// |  chrome.rdf URI parts      
const gksURNBase                 = "urn"
const gksURNBaseCont             = "mozilla";
const gksURNBasePack             = "package";
const gksURNBaseSkin             = "skin";
const gksURNBaseLoc              = "locale";
const gksURNBaseRoot             = "root";
const gksURNBasePacks            = "packages";
                                
// |  chrome.rdf node items     
const gksItemPack                = "package";
const gksItemBaseURL             = "baseURL";
const gksItemLocType             = "locType";
//const gksItemSelSkin             = "selectedSkin";
//const gksItemSelLoc              = "selectedLocale";

// |  rdf meta data and file paths
const gksMozillaNamespace        = "http://www.mozilla.org/rdf/chrome#";
const gkrsMozillaBaseCont        = [gksURNBase, gksURNBaseCont];
const gksMozillaXUL_Fastload     = "XUL";
const gksMozillaChromeRDF        = "chrome.rdf";
const gksMozillaInstChromeTxt    = "installed-chrome.txt";
const gksMozillaOverlayInfoDir   = "overlayinfo";
const gksMozillaOverlayInfo      = "overlayinfo/";
const gksMozillaOverlaysRDF      = "/content/overlays.rdf";
const gksMozillaStylesheetsRDF   = "/skin/stylesheets.rdf";
const gksMozillaOverlaysRoot     = "urn:mozilla:overlays"
const gksMozillaStylesheetsRoot  = "urn:mozilla:stylesheets"
const gksMozillaContentsRDF      = "contents.rdf";
const gksBackupExtension         = ".mnenhy.bak";
                                 
                                 
// |  god mode test              
const gksGodMode                 = "GodMode.txt";


// |  global prefs (see ReadPrefs)
var gbUninstallDebug             = null;  // dump debug info
var gbNoBackup                   = null;  // don't make backup
var gbNoShutdown                 = null;  // don't shut down


// |
// |  main functions
// |  
// |  Uninstall:                convenience function; will call several of the functions below
// |  UninstallPackages:        convenience function; will call several of the functions below
// |
// |  UninstallContentPackage:  uninstall the given content package from both user chrome
// |                            and app chrome (if possible) and remove all its files stored as
// |                            as baseURLs (if possible)
// |  UninstallSkinPackage:     as above, but for skin packages
// |  UninstallLocalePackage:   as above, but for locale packages
// |
// |
// |  BackupChromeRegistry:     backup all Mozilla system files possibly changed during uninstall
// |                            by renaming them to <name>.mnenhy.bak
// |                            user: chrome.rdf, overlayinfo
// |                            app:  chrome.rdf, overlayinfo, installed-chrome.txt
// |                            THIS FUNCTION SHOULD BE CALLED PRIOR TO ANY UNINSTALLING!
// |  RestoreChromeRegistry:    restore the files backupped via BackupChromeRegistry
// |
// |
// |  RemoveDeletableFiles:     delete package files that aren't used anymore
// |                            (this CANNOT be undone!)
// |  PurgeChromeRegistry:      delete XUL.mfasl, user overlayinfo and app overlayinfo,
// |                            rebuild both overlayinfo
// |  ShutdownMozilla:          shutdown Mozilla
// |


// Uninstall a package of the given type, but backup data before and clean up afterwards
// This function will shutdown Mozilla!
//  IN: array arPacks        -> an array of arrays of strings
//                              arName contains an entry for every package,
//                              every package is an array [package name, package type]
//                              (package type: see gksType*, defaults to gksTypeContent)
//                              eg. Uninstall([["foo", gksTypeSkin],["bar"]])
//                              will uninstall the skin package "foo" and the content package "bar"
function Uninstall(arPacks)
{
  debugdump("Mnenhy::Uninstall+\n");
  if (arPacks)
  { // make backup
    BackupChromeRegistry();
    // collect deletable files
    var oDeletableFiles = {};
    // uninstall by type
    UninstallPackages(arPacks, oDeletableFiles);
    // clean-up
    PurgeChromeRegistry();
    // shutdown and file deletion
    ShutdownMozilla(oDeletableFiles, true);
  }
  debugdump("Mnenhy::Uninstall-\n");
}


// Uninstall a package of the given type
//  IN:     array arPacks           -> an array of arrays of strings
//                                     arName contains an entry for every package,
//                                     every package is an array [package name, package type]
//                                     (package type: see gksType*, defaults to gksTypeContent)
//                                     eg. Uninstall([["foo", gksTypeSkin],["bar"]]) will uninstall 
//                                     the skin package "foo" and the content package "bar"
//  IN/OUT: object aoDeletableFiles -> collects the URIs that can be deleted by the caller
function UninstallPackages(arPacks, aoDeletableFiles)
{
  debugdump("Mnenhy::UninstallPackages+\n");
  if (arPacks)
  { // collect deletable files
    if (!aoDeletableFiles)
      aoDeletableFiles = {};
    // uninstall by type
    for (var iPack in arPacks)
    {
      const koPack = arPacks[iPack];
      const ksName = (koPack && (0 in koPack)) ? koPack[0] : "";
      const ksType = (koPack && (1 in koPack)) ? koPack[1] : "";

      switch(ksType)
      {
        case gksTypeSkin:   UninstallSkinPackage   (ksName, aoDeletableFiles); break;
        case gksTypeLocale: UninstallLocalePackage (ksName, aoDeletableFiles); break;
        case gksTypeContent:
        default:            UninstallContentPackage(ksName, aoDeletableFiles); break;
      }
    }
  }
  debugdump("Mnenhy::UninstallPackages-\n");
}


// Backup all Mozilla system files possibly changed during uninstall
// by renaming them to <name>.mnenhy.bak
//  user: chrome.rdf, overlayinfo
//  app:  chrome.rdf, overlayinfo, installed-chrome.txt
function BackupChromeRegistry()
{
  debugdump("Mnenhy::BackupChromeRegistry+");
  ReadPrefs();
  if (gbNoBackup)
    debugdump("Mnenhy::BackupChromeRegistry disabled!\n");
  else
  { // user files
    var oChromeDirUserBak = goMnenhy.m_koChromeDirUser.clone();
    oChromeDirUserBak.append(gksBackupExtension);
    goMnenhy.CopyFile(goMnenhy.m_koChromeDirUser,    gksMozillaChromeRDF,
                                  oChromeDirUserBak, gksMozillaChromeRDF,      true);
    goMnenhy.CopyFile(goMnenhy.m_koChromeDirUser,    gksMozillaOverlayInfoDir,
                                  oChromeDirUserBak, gksMozillaOverlayInfoDir, true);
    // app files
    if (this.mbIsGodMode)
    { // only if we're allowed to do so
      debugdump("Mnenhy::BackupChromeRegistry++\n");
      var oChromeDirAppBak = goMnenhy.m_koChromeDirApp.clone();
      oChromeDirAppBak.append(gksBackupExtension);
      goMnenhy.CopyFile(goMnenhy.m_koChromeDirApp,    gksMozillaChromeRDF,
                                    oChromeDirAppBak, gksMozillaChromeRDF,      true);
      goMnenhy.CopyFile(goMnenhy.m_koChromeDirApp,    gksMozillaOverlayInfoDir,
                                    oChromeDirAppBak, gksMozillaOverlayInfoDir, true);
      goMnenhy.CopyFile(goMnenhy.m_koChromeDirApp,    gksMozillaInstChromeTxt,
                                    oChromeDirAppBak, gksMozillaInstChromeTxt,  true);
      debugdump("Mnenhy::BackupChromeRegistry--\n");
    }
  }
  debugdump("Mnenhy::BackupChromeRegistry-\n");
}

// Restore the files backupped via BackupChromeRegistry
//  user: chrome.rdf, overlayinfo
//  app:  chrome.rdf, overlayinfo, installed-chrome.txt
function RestoreChromeRegistry()
{
  debugdump("Mnenhy::RestoreChromeRegistry+\n");
  // user files
  var oChromeDirUserBak = goMnenhy.m_koChromeDirUser.clone();
  oChromeDirUserBak.append(gksBackupExtension);
  goMnenhy.CopyFile(            oChromeDirUserBak, gksMozillaChromeRDF,
                    goMnenhy.m_koChromeDirUser,    gksMozillaChromeRDF,      true);
  goMnenhy.CopyFile(            oChromeDirUserBak, gksMozillaOverlayInfoDir,
                    goMnenhy.m_koChromeDirUser,    gksMozillaOverlayInfoDir, true);
  // app files
  if (this.mbIsGodMode)
  { // only if we're allowed to do so
    debugdump("Mnenhy::RestoreChromeRegistry++\n");
    var oChromeDirAppBak = goMnenhy.m_koChromeDirApp.clone();
    oChromeDirAppBak.append(gksBackupExtension);
    goMnenhy.CopyFile(            oChromeDirAppBak, gksMozillaChromeRDF,
                      goMnenhy.m_koChromeDirApp,    gksMozillaChromeRDF,      true);
    goMnenhy.CopyFile(            oChromeDirAppBak, gksMozillaOverlayInfoDir,
                      goMnenhy.m_koChromeDirApp,    gksMozillaOverlayInfoDir, true);
    goMnenhy.CopyFile(            oChromeDirAppBak, gksMozillaInstChromeTxt,
                      goMnenhy.m_koChromeDirApp,    gksMozillaInstChromeTxt,  true);
    debugdump("Mnenhy::RestoreChromeRegistry--\n");
  }
  debugdump("Mnenhy::RestoreChromeRegistry-\n");
}


// The members of the aoFiles object are URIs of the files to be deleted
function RemoveDeletableFiles(aoFiles, abBackup)
{
  debugdump("Mnenhy::RemoveDeletableFiles+\n");
  if (aoFiles)
  { // we need to release all cached jar files before deleting them
    // because Win2k (and maybe others) don't allow for deletion of open files
    const koJarProtHandler = goMnenhy.GetProtocolHandler(goMnenhy.ksProtocolJar);
    var oCache = null;
    if (koJarProtHandler instanceof Components.interfaces.nsIJARProtocolHandler)
      oCache = koJarProtHandler.JARCache;
    for (var sFile in aoFiles)
    {
      const koFile = ConvertURI2File(sFile);
      if (koFile.exists())
      { // release it
        try
        {
          if (oCache && koFile.isFile())
          {
            const koZipReader = oCache.getZip(koFile);
            if (koZipReader)
              koZipReader.close();
          }
          // if requested, make backup copies in the local .mnenhy.bak directory
          if (abBackup)
          { // insert backup dir into URI
            const koPH   = goMnenhy.GetProtocolHandler();
            var sURI     = koPH.getURLSpecFromFile(koFile);
            var sBakBase = koPH.getURLSpecFromFile(goMnenhy.m_koChromeDirUser);
            // first try profile
            if (sURI.search(sBakBase))
            {  // try app next
               sBakBase = koPH.getURLSpecFromFile(goMnenhy.m_koChromeDirApp);  
            }

            if (!sURI.search(sBakBase))
            { // create backup URI
              var sBakURI = sBakBase + "/" + gksBackupExtension + "/" + sURI.replace(sBakBase, "");
              const koBak = koPH.getFileFromURLSpec(sBakURI);
              debugdump("Mnenhy::RemoveDeletableFiles: Copying " + koFile.path + " to " + koBak.path + "\n")
              goMnenhy.CopyFile(koFile.parent, koFile.leafName, koBak.parent, koBak.leafName);
            }
            else
              dump("Mnenhy::RemoveDeletableFiles: Can't find base chrome dir in " + koFile.path + "\n");
          }
          debugdump("Mnenhy::RemoveDeletableFiles: Deleting " + koFile.path + "\n")
          koFile.remove(true);
        }
        catch(oError)
        {
          dump("Error in mnenhy-uninstall::remove deletable files(" + koFile.path + "):\n" 
               + oError + "\n");
          return oError;
        }
      } // found file
    } // loop over files
  } // have files
  debugdump("Mnenhy::RemoveDeletableFiles-\n");
}


/*
// The members of the aoFiles object are URIs of the files to be deleted
function RemoveDeletableFiles(aoFiles)
{ // Empty the nsIZipReaderCache of nsJar by faking a low memory condition
  // If we wouldn't do that, Win2k would block the deletion of files still in the cache
  debugdump("Mnenhy::RemoveDeletableFiles+\n");
  if (aoFiles)
  {
    const gksCID_ObserverService = "@mozilla.org/observer-service;1";
    var oObserverService = Components.classes[gksCID_ObserverService]
                                     .getService(Components.interfaces.nsIObserverService);
    if (oObserverService)
    { // send fake memory problem
      oObserverService.notifyObservers(null, "memory-pressure", "Mnenhy!");
      // delete the files
      for (var sFile in aoFiles)
      {
        const koFile = ConvertURI2File(sFile);
        if (koFile.exists())
        { // release it
          try
          {
            debugdump("Deleting " + koFile.path + "\n")
            koFile.remove(true);
          }
          catch(oError)
          {
            dump("Error in mnenhy-uninstall::remove deletable files(" + koFile.path + "):\n" 
                 + oError + "\n");
            return oError;
          }
        } // found file
      } // loop over files
    } // have service
  } // have files
  debugdump("Mnenhy::RemoveDeletableFiles-\n");
}
*/
  

// Delete XUL.mfasl, user overlayinfo and app overlayinfo, rebuild both overlayinfo
function PurgeChromeRegistry()
{ // user files
  debugdump("Mnenhy::PurgeChromeRegistry+\n");
  DeleteXulFastload();
  goMnenhy.DeleteFile(goMnenhy.m_koChromeDirUser, gksMozillaOverlayInfoDir);
  UpdateAllOverlayInfo(true);
  // app files
  if (this.mbIsGodMode)
  { // only if we're allowed to do so
    debugdump("Mnenhy::PurgeChromeRegistry++\n");
    goMnenhy.DeleteFile(goMnenhy.m_koChromeDirApp, gksMozillaOverlayInfoDir);
    UpdateAllOverlayInfo(false);
    debugdump("Mnenhy::PurgeChromeRegistry--\n");
  }
  debugdump("Mnenhy::PurgeChromeRegistry-\n");
}


// delete all deletable files, close all windows, shutdown Mozilla
function ShutdownMozilla(aoDeletableFiles, abBackup, asPrefsBranch)
{
  ReadPrefs();
  debugdump("Mnenhy::ShutdownMozilla+\n")

  // remove the remainder of the uninstalled packages
  CreateDeletableBaseURLs(mroChromeRDF[0], aoDeletableFiles);
  if (mroChromeRDF[1])      
    CreateDeletableBaseURLs(mroChromeRDF[1], aoDeletableFiles);

// Alternative method to call RemoveDeletableFiles
// (Not fully tested and not really better - the problem is goQuitApplication()!
//  Closing Mozilla via the window's upper right [x] button will do nicely, 
//  but goQuitApplication() makes it choke... :-| )
//
//const gksCID_ObserverService = "@mozilla.org/observer-service;1";
//var oObserverService = Components.classes[gksCID_ObserverService]
//                                 .getService(Components.interfaces.nsIObserverService);
//if (oObserverService)
//{
// 	var obstest = {observe: function(aoSubjectIgnored, asTopicIgnored, asStateIgnored) 
//                          {
//                            this.RemoveDeletableFiles(this.aoDeletableFiles, this.abBackup);
//                          },
//                 RemoveDeletableFiles: RemoveDeletableFiles,
//                 oDeletableFiles: aoDeletableFiles,
//                 bBackup: abBackup
//                 };
//  oObserverService.addObserver(obstest, "quit-application", false);
//}

  RemoveDeletableFiles(aoDeletableFiles, abBackup);
  
  // the following call to goQuitApplication will dump 
  // an obviously(?) bogus(?) error to the console
  if (!gbNoShutdown)
    goQuitApplication();
  debugdump("Mnenhy::ShutdownMozilla-\n")
}


// Uninstall a content package:
//  IN:     string asName           -> the name of the content package, eg. 'mnenhy'
//  IN/OUT: object aoDeletableFiles -> all deletable files are added as object members
//
// Uninstall a content package by
// - removing all skins/locales "urn:mozilla:<skin|locale>:<skin|lang>:<asName>"
//   incl. their installed-chrome.txt lines if applicable
//   (and adding their baseURLs to aoBaseURLs)
// - removing the content "urn:mozilla:package:<asName>"
//   incl. its installed-chrome.txt line if applicable
//   (and adding its baseURLs to aoBaseURLs)
// The files in aoBaseURLs can be deleted by the caller.
function UninstallContentPackage(asName, aoDeletableFiles)
{
  debugdump("Mnenhy::UninstallContentPackage("+asName+")+\n")
  if (asName && aoDeletableFiles)
  { // content pack URNs
    const krContent = [gksURNBasePack, asName];
    // loop over chrome.rdf files
    for (var iChrome = 0; (iChrome < mroChromeRDF.length) && (mroChromeRDF[iChrome]); ++iChrome)
    {
      var oChromeRDF   = mroChromeRDF[iChrome];
      var oElementData = {oChromeRDF: oChromeRDF,
                          sICText:    "",
                          bIsProfile: !iChrome,
                          oBaseURLs:  {} };
      var rNode;
      // open installed-chrome.txt (if in GodMode)
      if (this.mbIsGodMode)
        oElementData.sICText = LoadInstalledChrome();
      // process skin/locales
      const koSkinsAndLocales = oChromeRDF.GetSourcesForNode(gksItemPack, krContent);
      while (koSkinsAndLocales.hasMoreElements())
      { // read the baseURL from the skin or locale package entry before deleting it
        var oSkinOrLocale = koSkinsAndLocales.getNext()
        // get baseURL, delete installed-chrome.txt entry
        UninstallPackage(oSkinOrLocale, oElementData);
        // remove node, eg urn:mozilla:locale:de-DE:mnenhy
        rNode = oSkinOrLocale.Value.split(":").slice(2);
        oChromeRDF.RemoveNode(rNode);
        // if parent container is emtpy now, remove it, too
        rNode.pop();
        rNode.push(gksURNBasePacks);
        if (!oChromeRDF.GetContainerSize(rNode))
        { // remove container node, eg urn:mozilla:locale:de-DE:packages
          oChromeRDF.RemoveContainer(rNode);
          // and the containing <description>, eg. urn:mozilla:locale:de-DE
          rNode.pop();
          oChromeRDF.RemoveNode(rNode);
          // if this parent container is emtpy now also, remove it
          // (that does mean removing the skin/locale root!)
          rNode = [gksURNBasePack, gksURNBaseRoot];
          if (!oChromeRDF.GetContainerSize(rNode))
          { // remove the root node
            oChromeRDF.RemoveContainer(rNode);
          }
        }
      } // loop over skins and locales
      // process main element
      const koContent = oChromeRDF.ExistNode(krContent);
      if (koContent)
      { // remember content node baseURL
        UninstallPackage(koContent, oElementData);
        // remove node
        rNode = koContent.Value.split(":").slice(2);
        oChromeRDF.RemoveNode(rNode);
      }
      // save changes to installed-chrome.txt
      if (this.mbIsGodMode)
        SaveInstalledChrome(oElementData.sICText);
      // collect oBaseURLs
      for (var sBaseURL in oElementData.oBaseURLs)
        aoDeletableFiles[sBaseURL] = oElementData.oBaseURLs[sBaseURL];
    } // for loop over chrome.rdf files
  } // have args
  debugdump("Mnenhy::UninstallContentPackage-\n")
}


// Uninstall a skin or locale package:
//  IN:     string asProvider       -> the package type: gksURNBaseSkin or gksURNBaseLoc
//  IN:     string asName           -> the name of the package, eg. 'modern/1.0' or 'de-AT'
//  IN/OUT: object aoDeletableFiles -> deletable files will be added as object members
//
// Uninstall a skin or locale package by
// - deselecting it from all content packages
// - removing all skins/locales "urn:mozilla:<skin|locale>:<asName>:*"
//   incl. their installed-chrome.txt lines if applicable
//   (and adding their baseURLs to aoBaseURLs)
// - removing the skin/locale "urn:mozilla:<skin|locale>:<asName>"
//   incl. its installed-chrome.txt line if applicable
// The files in aoDeletableFiles can be deleted by the caller.
function UninstallSkinPackage(asName, aoDeletableFiles)
{
  debugdump("Mnenhy::UninstallSkinPackage+\n")
  UninstallSkinOrLocalePackage(gksURNBaseSkin, asName, aoDeletableFiles);
  debugdump("Mnenhy::UninstallSkinPackage-\n")
}
function UninstallLocalePackage(asName, aoDeletableFiles)
{
  debugdump("Mnenhy::UninstallLocalePackage+\n")
  UninstallSkinOrLocalePackage(gksURNBaseLoc, asName, aoDeletableFiles);
  debugdump("Mnenhy::UninstallLocalePackage-\n")
}
function UninstallSkinOrLocalePackage(asProvider, asName, aoDeletableFiles)
{
  debugdump("Mnenhy::UninstallSkinOrLocalePackage("+asProvider+","+asName+")+\n")
  var fDeselector = null;
  switch (asProvider)
  {
    case gksURNBaseSkin:  fDeselector = goMnenhy.m_koChromeRegService.deselectSkin;   break;
    case gksURNBaseLoc:   fDeselector = goMnenhy.m_koChromeRegService.deselectLocale; break;
  }

  if (fDeselector && asName && aoBaseURLs)
  { // skin/locale pack URNs
    const krSkinOrLoc = [asProvider, asName];
    // loop over chrome.rdf files
    for (var iChrome = 0; (iChrome < mroChromeRDF.length) && (mroChromeRDF[iChrome]); ++iChrome)
    {
      var oChromeRDF   = mroChromeRDF[iChrome];
      var oElementData = {oChromeRDF: oChromeRDF,
                          sICText:    "",
                          bIsProfile: !iChrome,
                          oBaseURLs:  {} };
      // open installed-chrome.txt (if in GodMode)
      if (this.mbIsGodMode)
        oElementData.sICText = LoadInstalledChrome();
      // try to deselect skin/locale
      try
      {
        fDeselector(asName, !iChrome);
      }
      catch(ignored){}
      // process skin/locales
      var oSkinOrLocPacks = krSkinOrLoc.slice();
      oSkinOrLocPacks.push(gksURNBasePacks);
      oChromeRDF.RemoveContainer(oSkinOrLocPacks, UninstallPackage, oElementData);
      // remove the skin/locales main <rdf:description>
      oChromeRDF.RemoveNode(krSkinOrLoc);
      // save changes to installed-chrome.txt
      if (this.mbIsGodMode)
        SaveInstalledChrome(oElementData.sICText);
      // collect oBaseURLs
      for (var sBaseURL in oElementData.oBaseURLs)
        aoDeletableFiles[sBaseURL] = oElementData.oBaseURLs[sBaseURL];
    } // loop over chrome.rdf files
  } // have valid arguments
  debugdump("Mnenhy::UninstallSkinOrLocalePackage-\n")
}


// Get the baseURL property of the nsIRDFResource aoElement and
// uninstall the corresponding installed-chrome.txt line (if possible).
// aoElementData must provide these members:
//    oChromeRDF: the chrome.rdf MnenhyRDF object to work with
//    oBaseURLs:  the storage member for the baseURLs
//    bIsProfile: is oChromeRDF the user profile chrome.rdf?
//    sICText:    the text content of the installed-chrome.txt
function UninstallPackage(aoElement, aoElementData)
{ // read the baseURL from the skin or locale package entry before deleting it
  debugdump("Mnenhy::UninstallPackage+\n")
  if (aoElement instanceof goMnenhy.rdf.gkIRDFResource)
  { // create node urn
    var rNode = aoElement.Value.split(":").slice(2);
    // read baseURL
    var sBaseURL = aoElementData.oChromeRDF.GetItemValue(gksItemBaseURL, rNode);
    if (sBaseURL)
    {
      aoElementData.oBaseURLs[sBaseURL] = aoElementData.bIsProfile;
      if (this.mbIsGodMode)
      { // delete installed-chrome.txt line (we don't know about path!)
        var sLine = CreateICLine(sBaseURL, rNode[0], aoElementData.bIsProfile, null, "\\s");
        var iSize = aoElementData.sICText.length;
        aoElementData.sICText = aoElementData.sICText.replace(new RegExp(sLine), "\n");
        if (aoElementData.sICText.length == iSize)
          dump("Mnenhy::UninstallPackage:Warning: couldn't find installed-chrome.txt line: "+sLine+"\n");
      } // app chrome
    }
  } // have nsIRDFResource
  debugdump("Mnenhy::UninstallPackage-\n")
}



// |
// |  uninstall helper functions
// |

// Check which of the baseURLs in the associative array aoBaseURLs
// are still in use in chrome.rdf aoChromeRDF.
// All baseURLs still in use will be deleted from aoBaseURLs.
// The elements of aoBaseURLs will be converted into nsIURI spec strings,
// nsIJARURIs will be replaced by their JARFile URIs.
// After runnning this functions, all files left in aoBaseURLs can be deleted!
function CreateDeletableBaseURLs(aoChromeRDF, aoBaseURLs)
{
  debugdump("Mnenhy::CreateDeletableBaseURLs+\n")
  if (aoChromeRDF && aoBaseURLs)
  { // prepare aoBaseURLs
    var iLength = 0;
    for (var sBaseURL in aoBaseURLs)
    {
      var oURI = goMnenhy.m_koIOService.newURI(sBaseURL, null, null);
      if (oURI instanceof Components.interfaces.nsIJARURI)
        oURI = oURI.JARFile;
      if (oURI.spec != sBaseURL)
      { // convert URL to URI
        aoBaseURLs[oURI.spec] = aoBaseURLs[sBaseURL];
        delete aoBaseURLs[sBaseURL];
      }
      ++iLength;
    }
    // check it!
    const koBaseURLRes = aoChromeRDF.GetItemRes(gksItemBaseURL);
    const koAllRes     = aoChromeRDF.moDataSource.GetAllResources();
    while (koAllRes.hasMoreElements() && iLength)
    { // inspect every resource
      const koRes = koAllRes.getNext();
      if (koRes instanceof goMnenhy.rdf.gkIRDFResource)
      { // check all baseURL arcs
        const koAllBaseURLs = aoChromeRDF.moDataSource.GetTargets(koRes, koBaseURLRes, true);
        while (koAllBaseURLs.hasMoreElements() && iLength)
        {
          const koBaseURL = koAllBaseURLs.getNext();
          if (koBaseURL instanceof goMnenhy.rdf.gkIRDFLiteral)
          { // using URIs makes JAR testing very easy
            var oURI = goMnenhy.m_koIOService.newURI(koBaseURL.Value, null, null);
            if (oURI instanceof Components.interfaces.nsIJARURI)
              oURI = oURI.JARFile;
            // check our baseURL array
            for (var sBaseURI in aoBaseURLs)
            {
              if (!oURI.spec.search(sBaseURI))
              { // still in use, mustn't delete that file
                delete aoBaseURLs[sBaseURI];
                --iLength;
              }
            } // for all aoBaseURLs
          } // baseURL is literal
        } // while all baseURLs
      } // resource is nsIRDFRresource
    } // while all resources
  } // valid parameters
  debugdump("Mnenhy::CreateDeletableBaseURLs-\n")
}


// loop over all packages in chrome.rdf and rebuild their overlayinfo
function UpdateAllOverlayInfo(abUser)
{
  debugdump("Mnenhy::UpdateAllOverlayInfo+\n")
  const koChromeRDF = mroChromeRDF[abUser ? 0 : 1]
  if (koChromeRDF)
  { // save and reload chrome.rdf
    koChromeRDF.FlushDataSource();

    // init the overlay cache
    // this will cache the open overlays.rdf and stylesheets.rdf
    // it's an associative array with path:object entries
    var oOverlayCache = {}
    
    // loop over content packages to generate overlayinfo
    var rsPack, sChromeURI, i;
    var rsPacks = koChromeRDF.GetNodeList([gksURNBasePack, gksURNBaseRoot]);
    if (rsPacks)
    {
      for (i in rsPacks)
      {
        rsPack = rsPacks[i].split(":").slice(2);
        if (rsPack)
        { // first read the package contents.rdf
          sChromeURI = koChromeRDF.GetItemValue(gksItemBaseURL, rsPack) + gksMozillaContentsRDF;
          UpdateOverlayInfo(oOverlayCache, sChromeURI);
        } // have urn
      } // for all content packages
    } // have content packages

    // loop over skin packages to generate overlayinfo
    const krsSkins = koChromeRDF.GetNodeList([gksURNBaseSkin, gksURNBaseRoot]);
    if (krsSkins)
    { // loop over all skin packages
      for (var j in krsSkins)
      { // get the content packages of this skin package
        const krsSkin = krsSkins[j].split(":").slice(2);
        const ksPacks = koChromeRDF.GetItemValue(gksURNBasePacks, krsSkin);
        if (ksPacks)
        {        
          const krsPacks = ksPacks.split(":").slice(2);
          rsPacks = koChromeRDF.GetNodeList(krsPacks);
          // loop over all packages of this skin
          for (i in rsPacks)
          {
            rsPack = rsPacks[i].split(":").slice(2);
            if (rsPack)
            { // first read the package contents.rdf
              sChromeURI = koChromeRDF.GetItemValue(gksItemBaseURL, rsPack) + gksMozillaContentsRDF;
              UpdateOverlayInfo(oOverlayCache, sChromeURI);
            } // have urn
          } // for all skin content packages
        } // have skin content packages
      } // for all skins          
    } // have skin packages      

    // save all files in cache to disk
    for (i in oOverlayCache)
    {
      oOverlayCache[i].FlushDataSource(oOverlayCache[i].msDataSourceURI);
      delete oOverlayCache[i];
    }
  } // have chrome.rdf          
  debugdump("Mnenhy::UpdateAllOverlayInfo-\n");
}


// update the overlayinfo with the information of the overlays and
// stylesheets sections of the contents.rdf given by asContentsURI
// (this function contains one subfunction!)
function UpdateOverlayInfo(aoOverlayCache, aoContentsURI)
{ // mark this overlayinfo recreation
  debugdump("Mnenhy::UpdateOverlayInfo("+aoContentsURI+")+\n");
  if (aoOverlayCache && aoContentsURI)
  { // init the contents.rdf file to be read
    if (typeof(aoContentsURI) == "string")
      aoContentsURI = goMnenhy.m_koIOService.newURI(aoContentsURI, null, null);
    const koContentsRDF = new goMnenhy.rdf.MnenhyRDF(aoContentsURI, null, null, null, true);
    if (koContentsRDF.IsGood())
    { // loop over overlays and stylesheets
      const krsRootSeqs = [gksMozillaOverlaysRoot, gksMozillaStylesheetsRoot];
      for (var sRootSeq in krsRootSeqs)
      { // process the root <Seq> and write its data into the respective overlayinfo file
        const koOuterSeq = koContentsRDF.InitContainer(krsRootSeqs[sRootSeq]);
        if (koOuterSeq)
        { // root <Seq> found
          debugdump(" Mnenhy::UpdateOverlayInfo: RootSeq="+koOuterSeq.Resource.Value+"\n");
          const kbIsOverlay = (sRootSeq == "0");
          const koOuterEnum = koOuterSeq.GetElements();
          while (koOuterEnum.hasMoreElements())
          { // get source container
            const koSeq      = koOuterEnum.getNext();
            const koInnerSeq = koContentsRDF.InitContainer(koSeq);
            debugdump("  Mnenhy::UpdateOverlayInfo: koInnerSeq="+koInnerSeq.Resource.Value+"\n")
            // get handle to target overlayinfo file
            var oInfoRDF = null;
            const ksInfoFileURI = GetOverlayInfoFile(koContentsRDF.msDataSourceURI, koSeq.Value, kbIsOverlay);
            debugdump("  Mnenhy::UpdateOverlayInfo: ksInfoFileURI="+ksInfoFileURI+"\n")
            if (!(ksInfoFileURI in aoOverlayCache))
            { // open new overlayinfo file
              oInfoRDF = new goMnenhy.rdf.MnenhyRDF(ksInfoFileURI, null, null, null, true);
              if (oInfoRDF)
              { // remember the RDF file 
                aoOverlayCache[ksInfoFileURI] = oInfoRDF;
                // if we haven't touched it yet, empty it
                debugdump("  Mnenhy::UpdateOverlayInfo: emptying ksInfoFileURI\n");
                const koAllRes = oInfoRDF.moDataSource.GetAllResources();
                while (koAllRes.hasMoreElements())
                { // inspect every resource
                  const koRes = koAllRes.getNext();
                  oInfoRDF.RemoveNode(koRes);
                } // while all resources
              }
            }
            if (ksInfoFileURI in aoOverlayCache)
            { // get RDF file
              if (!oInfoRDF)
                oInfoRDF = aoOverlayCache[ksInfoFileURI];
              // create target container
              const koSourceSeq = koInnerSeq.Resource;
              var oTargetSeq = oInfoRDF.InitContainer(koSourceSeq);
              if (!oTargetSeq)
              { // create new Seq
                oInfoRDF.AddContainer(koSourceSeq);
                oTargetSeq = oInfoRDF.InitContainer(koSourceSeq);
              }
              if (oTargetSeq)
              { // have target <Seq>
                const koInnerEnum = koInnerSeq.GetElements();
                while (koInnerEnum.hasMoreElements())
                { // get source container
                  const koLit = koInnerEnum.getNext();
                  if (koLit instanceof goMnenhy.rdf.gkIRDFLiteral)
                  {                
                    debugdump("   Mnenhy::UpdateOverlayInfo: koLit="+koLit.Value+"\n")
                    // set target literal
                    if (oTargetSeq.IndexOf(koLit) == -1)
                      oTargetSeq.AppendElement(koLit);
                  } // have literal
                } // loop over overlayinfo entries
              } // have target <Seq>
            } // have overlayinfo file
          } // loop over all overlays/stylesheets
        } // root <Seq> found
      } // for overlays/stylesheets
    } // have valid aoContentsURI  
  }
  debugdump("Mnenhy::UpdateOverlayInfo-\n");
}


// |
// |  helper functions
// |  (actually not really related to uninstall, but needed ;-) )
// |

// read all the prefs needed in this file
function ReadPrefs()
{
  debugdump("Mnenhy::ReadPrefs+\n");
  gbNoBackup       = goMnenhy.GetPref(gksUninstallNoBackupPref,   "bool");
  debugdump("Mnenhy::ReadPrefs: gbNoBackup      ="+gbNoBackup      +"\n");
  gbNoShutdown     = goMnenhy.GetPref(gksUninstallNoShutdownPref, "bool");
  debugdump("Mnenhy::ReadPrefs: gbNoShutdown    ="+gbNoShutdown    +"\n");
  gbUninstallDebug = goMnenhy.GetPref(gksUninstallDebugPref,      "bool");
  debugdump("Mnenhy::ReadPrefs: gbUninstallDebug="+gbUninstallDebug+"\n");
  debugdump("Mnenhy::ReadPrefs-\n");
}


// Do we have administra(i)tor rights for the Mozilla install chrome directory?
//
// return value:
//    true:     we are root!
//    false:    we're in user hell.
function IsGodMode()
{ // try to write a file into the install chrome directory and delete it
  debugdump("Mnenhy::IsGodMode+\n")
  var bGod = false;
  try
  { // create file
    const klGodPermissions = 0644;  // only owner may change, others only read
    var oGodFile = goMnenhy.m_koChromeDirApp.clone();
    oGodFile.append(gksGodMode)
    oGodFile.create(oGodFile.NORMAL_FILE_TYPE, klGodPermissions);
    // delete the file
    oGodFile.remove(false);
    bGod = true;
  }
  catch(ignored)
  {
    debugdump(ignored+"\n")
  }
  debugdump("Mnenhy::IsGodMode-: "+bGod+"\n")
  return bGod;
}


// get the chrome.rdf
// if aoPath is a non-null nsIFile, look there
// else try the user profile dir (abUser = true) or the app chrome dir (abUser = false).
function GetChromeRDF(abUser, aoPath)
{
  debugdump("Mnenhy::GetChromeRDF+\n")
  var oRDF  = null;
  // get the path
  var oPath = aoPath;
  if (!oPath)
    oPath = abUser ? goMnenhy.m_koChromeDirUser : goMnenhy.m_koChromeDirApp;
  // get the chrome.rdf file
  var oChrome = oPath.clone();
  if (oChrome)
  { // open it as RDF
    oChrome.append(gksMozillaChromeRDF);
    oChrome = goMnenhy.GetProtocolHandler().newFileURI(oChrome);
    oRDF = new goMnenhy.rdf.MnenhyRDF(oChrome, null, gksMozillaNamespace, gkrsMozillaBaseCont);
  }
  debugdump("Mnenhy::GetChromeRDF-\n")
  return oRDF;
}


// create a line for use in installed-chrome.txt
// uri must not be empty; defaults for the other parametres are content, install, url, LF
// asLocation accepts |true| as an alias for gksLocProfile
function CreateICLine(asURI, asProvider, asLocation, asURIType, asLF)
{ // very thin ice indeed ;-)
  debugdump("Mnenhy::CreateICLine("+asURI+", "+asProvider+", "+asLocation+", "+asURIType+", "+asLF+")+\n")
  var sLine = null;
  if (asURI)             
  {
    // default provider is content
    switch (asProvider)
    {
      case gksTypeContent:
      case gksTypeSkin:   
      case gksTypeLocale: 
        break;
      default:
        asProvider = gksTypeContent;
    }
    // default location is application
    switch (asLocation)
    {
      case gksLocInstall:
      case gksLocProfile:    
      case gksLocSelect:
        break;
      case true:
        asLocation = gksLocProfile;
        break;
      default:
        asLocation = gksLocInstall;
    }
    // default URI type is url
    switch (asURIType)
    {
      case gksURIURL:
      case gksURIPath:    
        break;
      default:
        asURIType = gksURIURL;
    }
    // default LF is LF  
    if (!asLF)
      asLF = gksLF
    // create  
    sLine = asProvider + "," + asLocation + "," + asURIType + "," + asURI + asLF;
  }
  debugdump("Mnenhy::CreateICLine-\n")
  return sLine;
}


// get the installed-chrome.txt as a nsIFile
function GetInstalledChromeFile()
{
  debugdump("Mnenhy::GetInstalledChromeFile+\n")
  var oFile = goMnenhy.m_koChromeDirApp.clone();
  oFile.append(gksMozillaInstChromeTxt);
  debugdump("Mnenhy::GetInstalledChromeFile-\n")
  return oFile;
}
// get the contents of installed-chrome.txt
function LoadInstalledChrome()
{
  debugdump("Mnenhy::LoadInstalledChrome\n")
  return goMnenhy.ReadFile(GetInstalledChromeFile());
}
// save sContent to installed-chrome.txt
function SaveInstalledChrome(sContent)
{ // clean-up: remove all CR, remove whitespace lines, keep only one LF, expand LF to CRLF
  debugdump("Mnenhy::SaveInstalledChrome+\n")
  if (!sContent)
  {
    sContent = "";
  }
  else
  {
    sContent = sContent.replace(/\r/g,    "");
    sContent = sContent.replace(/\s*\n/g, gksLF);
    sContent = sContent.replace(/(\n)+/g, gksLF);
    sContent = sContent.replace(/\n/g,    gksCRLF);
    sContent += gksCRLF;
  }
  // save
  goMnenhy.WriteFile(GetInstalledChromeFile(), sContent);
  debugdump("Mnenhy::SaveInstalledChrome-\n")
}


// get the overlayinfo file URI string for either overlays or stylesheets
// aoSourceURI can be of type string or nsIURI
function GetOverlayInfoFile(aoSourceURI, aoTargetURI, abOverlays)
{
  debugdump("Mnenhy::GetOverlayInfoFile+\n")
  var sInfoFile = "";
  if (aoSourceURI && aoTargetURI)
  {
    if (typeof(aoTargetURI) == "string")
      aoTargetURI = goMnenhy.m_koIOService.newURI(aoTargetURI, null, null);

    const koFPH  = goMnenhy.GetProtocolHandler();
    const ksUser = goMnenhy.ConvertURI2FileURI(koFPH.newFileURI(goMnenhy.m_koChromeDirUser)).spec;
    const ksApp  = goMnenhy.ConvertURI2FileURI(koFPH.newFileURI(goMnenhy.m_koChromeDirApp )).spec;

    var sFileURI = goMnenhy.ConvertURI2FileURI(aoSourceURI);
    if (sFileURI)
    {
      var sInfoURI = "";
      sFileURI = sFileURI.spec;

      // get the location (user or app)
      if (!sFileURI.search(ksUser))
      { // we have a user chrome uri
        sInfoURI = ksUser;
        sFileURI = sFileURI.replace(ksUser, "");
      }
      else if (!sFileURI.search(ksApp))
      { // we have a app chrome uri
        sInfoURI = ksApp;
        sFileURI = sFileURI.replace(ksApp, "");
      }

      // get the package name
      if (sInfoURI && aoTargetURI.schemeIs("chrome"))
      { // canonify the chrome uri
        goMnenhy.m_koChromeRegService.canonify(aoTargetURI);
        // get the package name
        const krPack = aoTargetURI.spec.match("^chrome://(.*?)(?=\.jar|/)");
        if (1 in krPack)
        { // found package name
          sInfoURI += gksMozillaOverlayInfo
                      + krPack[1]
                      + (abOverlays ? gksMozillaOverlaysRDF : gksMozillaStylesheetsRDF);
          sInfoFile = sInfoURI;
        } // found package name
      } // if (sInfoURI)
    } // have sFileURI
  }
  debugdump("Mnenhy::GetOverlayInfoFile-\n")
  return sInfoFile;
}


// create a nsIFile from the given aoURI (given as string or nsIURI)
function ConvertURI2File(aoURI)
{
  debugdump("Mnenhy::ConvertURI2File+\n")
  var oFile = null;
  if (aoURI)
  {
    const koFileURI = goMnenhy.ConvertURI2FileURI(aoURI);
    if (koFileURI)
    { // create nsIFile from URI
      oFile = koFileURI.file.clone();
    }
  }
  debugdump("Mnenhy::ConvertURI2File-\n")
  return oFile;
}


// remove $UserProfileDir/xul.mfl (resp. xul.mfasl)
function DeleteXulFastload()
{
  debugdump("Mnenhy::DeleteXulFastload+\n")
  try
  {
    var oFaslService = Components.classes[gksCID_Fastload]
                                 .getService(Components.interfaces.nsIFastLoadService);
    var oFaslFile    = oFaslService.newFastLoadFile(gksMozillaXUL_Fastload);
    oFaslFile.remove(false);
  }
  catch(ignored){}
  debugdump("Mnenhy::DeleteXulFastload-\n")
}


// debug
function debugdump(aoDump)
{
  if (gbUninstallDebug)
    dump(aoDump);
}


// |
// |  Init called by the global Mnenhy object when loading this script file
// |
function Init(aoMnenhy)
{ // keep pointer to have the Mnenhy object prototype available
  this.goMnenhy     = aoMnenhy;
  // get certain flags
  ReadPrefs();
  // allow app chrome modifications?
  this.mbIsGodMode  = IsGodMode();
  // chrome RDF files
  this.mroChromeRDF = [GetChromeRDF(true), (mbIsGodMode ? GetChromeRDF(false) : null)];
  // get the shutdown function into the hidden window
  goMnenhy.Include("chrome://global/content/globalOverlay.js", this);  
}